#ifdef WIN32
#pragma comment(lib, "c:\\TopfieldDisk\\DataFiles\\SDK\\TAP_SDK_DLL")
#include "TAP_Emulations.h"
#endif

#include "tap.h"
#include "menu.h"
#include "dlmalloc.h"

void saveAutoDeleteList();
bool isRec(char *name);
		
extern char log_buffer[512];
extern char curr_dir[256];
extern char file_name[100];
extern int no_of_auto_deletes;
extern dword recurse_subfolders;

bool InDataFilesFolder(char* path);
bool ChangeDirPath(char *path);
char* GetCurrentDir(void);

extern struct _auto_delete_list
{
	char file_name[100];
	word delete_after_days;
	word delete_after_min;
	word delete_after_max;
	word delete_when_disk;
	word spare_1;
	word spare_2;
	word spare_3;
	word spare_4;
} *auto_delete_list[512], auto_delete_copy, *auto_delete_switch;


/* load the list in from file, auto_delete.dat */
void loadAutoDeleteList()
{
	char dat_name[20], *load_dump;
	TYPE_File *file;
	word i;
	
	ChangeDirMediaManager();

	strcpy(dat_name, "auto_delete.dat");
	no_of_auto_deletes = 0;

	if (TAP_Hdd_Exist(dat_name))
	{
		if (verbose) if (write_log) logText("Reading auto delete list from file", FALSE);
		file = TAP_Hdd_Fopen(dat_name);
		if (file != NULL)
		{
			TAP_Hdd_Fread(&no_of_auto_deletes, sizeof(int), 1, file);

			if (no_of_auto_deletes > 0) {
				load_dump = dlmalloc(sizeof(*auto_delete_list[0]) * no_of_auto_deletes);
				TAP_Hdd_Fread(load_dump, sizeof(*auto_delete_list[0]), no_of_auto_deletes, file);
				for (i = 0; i < no_of_auto_deletes; i++) {
					auto_delete_list[i] = dlmalloc(sizeof(*auto_delete_list[i]));
					memcpy(auto_delete_list[i], load_dump, sizeof(*auto_delete_list[i]));
					load_dump += sizeof(*auto_delete_list[i]);
				}

				load_dump -= sizeof(*auto_delete_list[0]) * no_of_auto_deletes;
				dlfree(load_dump);
				dlmalloc_trim(0);
			}

			//for (i = 0; i < no_of_auto_deletes; i++) {
			//	auto_delete_list[i] = dlmalloc(sizeof(*auto_delete_list[i]));
			//	TAP_Hdd_Fread(auto_delete_list[i], sizeof(*auto_delete_list[i]), 1, file);
			//}

			TAP_Hdd_Fclose(file);
		}
	}

	ChangeDirPath(curr_dir);
}

void appendToAutoDeleteList(char *file_name)
{
	if (no_of_auto_deletes >= 512)
		return ;

	auto_delete_list[no_of_auto_deletes] = dlmalloc(sizeof(*auto_delete_list[no_of_auto_deletes]));
	strcpy(auto_delete_list[no_of_auto_deletes]->file_name, file_name);
	auto_delete_list[no_of_auto_deletes]->delete_after_days = 0;
	auto_delete_list[no_of_auto_deletes]->delete_after_min = 0;
	auto_delete_list[no_of_auto_deletes]->delete_after_max = 0;
	auto_delete_list[no_of_auto_deletes]->delete_when_disk = 0;
	auto_delete_list[no_of_auto_deletes]->spare_1 = 0;
	auto_delete_list[no_of_auto_deletes]->spare_2 = 0;
	auto_delete_list[no_of_auto_deletes]->spare_3 = 0;
	auto_delete_list[no_of_auto_deletes]->spare_4 = 0;
	no_of_auto_deletes++;
	//saveAutoDeleteList();
}

void removeFromAutoDeleteList(int delete_idx)
{
	int i;

	// shuffle up auto deletes and delete last entry
	for (i = delete_idx; i < (no_of_auto_deletes-1); i++)
	{
		strcpy(auto_delete_list[i]->file_name, auto_delete_list[i+1]->file_name);
		auto_delete_list[i]->delete_after_days = auto_delete_list[i+1]->delete_after_days;
		auto_delete_list[i]->delete_after_min = auto_delete_list[i+1]->delete_after_min;
		auto_delete_list[i]->delete_after_max = auto_delete_list[i+1]->delete_after_max;
		auto_delete_list[i]->delete_when_disk = auto_delete_list[i+1]->delete_when_disk;
		auto_delete_list[i]->spare_1 = auto_delete_list[i+1]->spare_1;
		auto_delete_list[i]->spare_2 = auto_delete_list[i+1]->spare_2;
		auto_delete_list[i]->spare_3 = auto_delete_list[i+1]->spare_3;
		auto_delete_list[i]->spare_4 = auto_delete_list[i+1]->spare_4;
	}
	dlfree(auto_delete_list[no_of_auto_deletes-1]);
	no_of_auto_deletes--;
}

// remove the last entry
void rollbackAutoDeleteList()
{
	dlfree(auto_delete_list[no_of_auto_deletes-1]);
	no_of_auto_deletes--;
}

void removeAllFromAutoDeleteList()
{
	int i;

	for (i = 0; i < no_of_auto_deletes; i++)
	{
		dlfree(auto_delete_list[i]);
	}
	no_of_auto_deletes = 0;
}

int existsInAutoDeleteList(char *file_name)
{
	int i;

	for (i = 0; i < no_of_auto_deletes; i++)
	{
		if (strcmp(file_name, auto_delete_list[i]->file_name) == 0)
			return i;
	}

	return -1;
}

/* save a copy of the auto delete rules for later restore */
void copyAutoDeleteEntry(int copy_idx)
{
	auto_delete_copy = *auto_delete_list[copy_idx];
}

/* restore auto delete rules for specified idx */
void restoreAutoDeleteEntry(int restore_idx)
{
	strcpy(auto_delete_list[restore_idx]->file_name, auto_delete_copy.file_name);
	auto_delete_list[restore_idx]->delete_after_days = auto_delete_copy.delete_after_days;
	auto_delete_list[restore_idx]->delete_after_min = auto_delete_copy.delete_after_min;
	auto_delete_list[restore_idx]->delete_after_max = auto_delete_copy.delete_after_max;
	auto_delete_list[restore_idx]->delete_when_disk = auto_delete_copy.delete_when_disk;
	auto_delete_list[restore_idx]->spare_1 = auto_delete_copy.spare_1;
	auto_delete_list[restore_idx]->spare_2 = auto_delete_copy.spare_2;
	auto_delete_list[restore_idx]->spare_3 = auto_delete_copy.spare_3;
	auto_delete_list[restore_idx]->spare_4 = auto_delete_copy.spare_4;
}

void moveAutoDeleteEntryDown(int shuffle_idx)
{
	if (shuffle_idx >= no_of_auto_deletes - 1)
		return;

	auto_delete_switch = auto_delete_list[shuffle_idx+1];

	// swap playlist entry with one above
	auto_delete_list[shuffle_idx+1] = auto_delete_list[shuffle_idx];
	auto_delete_list[shuffle_idx] = auto_delete_switch;
}

void moveAutoDeleteEntryUp(int shuffle_idx)
{
	if (shuffle_idx <= 0)
		return;

	auto_delete_switch = auto_delete_list[shuffle_idx-1];

	// swap playlist entry with one above
	auto_delete_list[shuffle_idx-1] = auto_delete_list[shuffle_idx];
	auto_delete_list[shuffle_idx] = auto_delete_switch;
}

void saveAutoDeleteList()
{
	TYPE_File *file;
	word i;
	char *save_dump;

	// save auto delete list to file
	ChangeDirMediaManager();

	TAP_Hdd_Delete("auto_delete.dat");
	TAP_Hdd_Create("auto_delete.dat", ATTR_NORMAL);
	file = TAP_Hdd_Fopen("auto_delete.dat");
	TAP_Hdd_Fwrite(&no_of_auto_deletes, sizeof(int), 1, file);

	if (no_of_auto_deletes > 0) {
		save_dump = dlmalloc(sizeof(*auto_delete_list[0]) * no_of_auto_deletes);

		for (i = 0; i < no_of_auto_deletes; i++) {
			memcpy(save_dump, auto_delete_list[i], sizeof(*auto_delete_list[i]));
			save_dump += sizeof(*auto_delete_list[i]);
		}
		save_dump -= sizeof(*auto_delete_list[0]) * no_of_auto_deletes;
		TAP_Hdd_Fwrite(save_dump, sizeof(*auto_delete_list[i]), no_of_auto_deletes, file);
		
		dlfree(save_dump);
		dlmalloc_trim(0);
	}

	//for (i = 0; i < no_of_auto_deletes; i++)
	//	TAP_Hdd_Fwrite(auto_delete_list[i], sizeof(*auto_delete_list[i]), 1, file);

	TAP_Hdd_Fclose(file);

	ChangeDirPath(curr_dir);
}