#ifdef WIN32
#pragma comment(lib, "c:\\TopfieldDisk\\DataFiles\\SDK\\TAP_SDK_DLL")
#include "TAP_Emulations.h"
#endif

#include "tap.h"
#include "dlmalloc.h"

#define LE16( p )	( (((p&0x00ff))<<8) | ((p&0xff00)>>8) )
#define LE32( p )	( (((p&0x000000ff))<<24) | (((p&0x0000ff00))<<8) | (((p&0x00ff0000))>>8) | ((p&0xff000000)>>24) )

struct BMP_HEAD {
    byte id[2];
	word size_l;
	word size_h;
    word reserved[2];
    word offset_l;
	word offset_h;
};

struct BMP_INFO {
    long info_size;
    long width;
    long height;
    word plane;
    word bits_per_pixel;
    long compress;
    long img_size;
    long x_res;
    long y_res;
    long color;
    long icolor;
};

static struct BMP_HEAD head;
static struct BMP_INFO info;

static int capture_num = 0;
static char out_name[64];

TYPE_File *bitmap_file;

#define _R(y,u,v) (0x2568*(y)			   + 0x3343*(u)) /0x2000
#define _G(y,u,v) (0x2568*(y) - 0x0c92*(v) - 0x1a1e*(u)) /0x2000
#define _B(y,u,v) (0x2568*(y) + 0x40cf*(v)) 			 /0x2000

#define _S(a)		(a)>255 ? 255 : (a)<0 ? 0 : (a)

/*void convertBitmapToPixels(word *target, byte *source, int width, int height)
{
	word pixel, y;
	int read_idx, target_idx = 0;

	for (y = height - 1; y > 0; y--)
	{
		read_idx = 0;

		while (read_idx < ((width * 3) - 3))
		{
			// colour pixel made up of three different value for RGB
			if ((source+(y*width)+read_idx+2) == NULL || (source+(y*width)+read_idx+1) == NULL || (source+(y*width)+read_idx) == NULL)
				return;

			pixel = RGB8888(*(source+(y*width)+read_idx+2), *(source+(y*width)+read_idx+1), *(source+(y*width)+read_idx));

			// save pixel in the cache memory space
			target[target_idx] = pixel;

			target_idx++;
			read_idx += 3;
		}
	}
}*/

void convertBitmapToPixels(word *target, byte *source, int width, int height)
{
	int x, y;
	byte r, g, b;
	byte temp[3];

	if (!target || !source)
		return;

	for (y = height-1; y >= 0; --y) {
        for (x = 0; x < width; ++x) {
            // This seems to be the actual format 
			memcpy(temp, source, 3);
            r = temp[2];
            g = temp[1];
            b = temp[0];
			target[x+y*width] = RGB8888(r, g, b);
			source += 3;
        }

		// Chomp to the next 32-bit boundary 
        if ((x % 4) != 0)
			source = source + (x % 4);
    }
}

void YUV2RGB(word yy, word uu, word vv, byte *r, byte *g, byte *b)
{
	signed int _r,_g,_b;
	signed int y, u, v;

	y = yy;
	u = (int)uu - 128;
	v = (int)vv - 128;

	_r = _R(y,u,v);
	_g = _G(y,u,v);
	_b = _B(y,u,v);

	*r = _S(_r);
	*g = _S(_g);
	*b = _S(_b);
}

static dword rev_endian(dword x, int bytes)
{
    if (bytes==1) return x;
    if (bytes==2) return ((x>>8)&0xFF) | ((x<<8)&0xFF00);
    if (bytes==3) return ((x>>16)&0xFF) | ((x)&0xFF00) | ((x<<16)&0xFF0000);
    return ((x>>24)&0xFF) | ((x>>8)&0xFF00) | ((x<<8) & 0xFF0000) | ((x<<24)&0xFF000000);
}

static dword endian2(byte u1, byte u2, int endian)
{
    if (endian)
      return ((u1<<8)&0xFF00) | ((u2)&0xFF);
    else
      return ((u2<<8)&0xFF00) | ((u1)&0xFF);
}

static dword endian4(byte u1, byte u2, byte u3, byte u4, int endian)
{
    if (endian)
      return ((u1<<24)&0xFF000000) | ((u2<<16)&0xFF0000)
           | ((u3<<8)&0xFF00) | ((u4)&0xFF);
    else
      return ((u4<<24)&0xFF000000) | ((u3<<16)&0xFF0000)
           | ((u2<<8)&0xFF00) | ((u1)&0xFF);
}

static int fread_endian(dword *x, int bytes, TYPE_File *fp, int endian)
{
    byte u1, u2, u3, u4;
    if (bytes==1) {
        if (TAP_Hdd_Fread(&u1, 1, 1, fp) != 1) return -1;
        *x = u1;
        return 1;
    }
    else if (bytes==2) {
        if (TAP_Hdd_Fread(&u1, 1, 1, fp) != 1) return -1;
        if (TAP_Hdd_Fread(&u2, 1, 1, fp) != 1) return -1;
        *x = endian2(u1, u2, endian);
        return 2;
    }
    else if (bytes==4) {
        if (TAP_Hdd_Fread(&u1, 1, 1, fp) != 1) return -1;
        if (TAP_Hdd_Fread(&u2, 1, 1, fp) != 1) return -1;
        if (TAP_Hdd_Fread(&u3, 1, 1, fp) != 1) return -1;
        if (TAP_Hdd_Fread(&u4, 1, 1, fp) != 1) return -1;
        *x = endian4(u1, u2, u3, u4, endian);
        return 4;
    }
    else {
        *x = 0;
        return 0;
    }
}

word *loadAndDecodeBitmap(char *name, int *width, int *height)
{
	byte *bitmap_data = NULL;
	word *target_data = NULL;

    int Endian = 1;
    byte u1tmp, u2tmp;
    dword ULtmp;
    dword fileSize;
    dword dataOffset;
    dword hdrVersion;
    dword nPlanes;
    dword bitCount;
    dword compression;
    dword imageSize;
    dword colorsUsed;
    dword colorsImportant;
    //byte colorTable[256][4];

	// open the bitmap file
	TAP_Print("Open bmp %s\r\n", name);
	bitmap_file = TAP_Hdd_Fopen(name);
	if (bitmap_file == NULL)
		return NULL;

	// read in the header structures
	//TAP_Hdd_Fread(&head, sizeof(head), 1, bitmap_file);
	//TAP_Hdd_Fread(&info, sizeof(info), 1, bitmap_file);

	TAP_Hdd_Fread(&u1tmp, 1, 1, bitmap_file);
    TAP_Hdd_Fread(&u2tmp, 1, 1, bitmap_file);
	TAP_Print("Id %c%c\r\n", u1tmp, u2tmp);
	if (u1tmp != 66 || u2tmp != 77) { /* 'B' and 'M' in order */
		TAP_Hdd_Fclose(bitmap_file);
		return NULL;
	}

    fread_endian(&fileSize, 4, bitmap_file, Endian);

    /* Skip 4-byte reserved segment */
    TAP_Hdd_Fseek(bitmap_file, 4, 1);

    fread_endian(&dataOffset, 4, bitmap_file, Endian);

    /* These 4 bytes hold the size of the info header.
     * Pre-standard DIB images have 12-byte "CORE" headers.
     * Standard .BMP files have 40-byte headers.
     * 'Version 4' BMPs have 108-byte headers.
     * 'Version 5' BMPs have 124-byte headers.
     * We don't support the last two formats; hardly anyone does!
     */
    fread_endian(&hdrVersion, 4, bitmap_file, Endian);

    if (hdrVersion > 0x10000) {
        /* No header is that big!  Swap endianness and try again. */
        Endian = !Endian;
        fileSize = rev_endian(fileSize, 4);
        dataOffset = rev_endian(dataOffset, 4);
        hdrVersion = rev_endian(hdrVersion, 4);
    }
	TAP_Print("Header size %d\r\n", hdrVersion);

    if (hdrVersion == 12) {
        /* Read a DIB image. */
        fread_endian(&ULtmp, 2, bitmap_file, Endian);
        *width = 30000 & ULtmp;
        fread_endian(&ULtmp, 2, bitmap_file, Endian);
        *height = 30000 & ULtmp;
        fread_endian(&nPlanes, 2, bitmap_file, Endian);
        fread_endian(&bitCount, 2, bitmap_file, Endian);

        /* Assume the following: */
        compression = 0;
        imageSize = (*width)*(*height);
        if (bitCount == 4) imageSize /= 2;
        else if (bitCount == 24) imageSize *= 3;
        else if (bitCount == 32) imageSize *= 4;
		TAP_Print("Bit count %d\r\n", bitCount);
        colorsUsed = 0;
    }
	else if (hdrVersion == 40) {
		fread_endian(&ULtmp, 4, bitmap_file, Endian);
		*width = (int) ULtmp;
		fread_endian(&ULtmp, 4, bitmap_file, Endian);
		*height = (int) ULtmp;

		fread_endian(&nPlanes, 2, bitmap_file, Endian);
		fread_endian(&bitCount, 2, bitmap_file, Endian);
		TAP_Print("Bit count %d\r\n", bitCount);

		if (bitCount != 1 && bitCount != 4 && bitCount != 8
						  && bitCount != 16 && bitCount != 24
						  && bitCount != 32) {
			TAP_Hdd_Fclose(bitmap_file);
			return NULL;
		}

		fread_endian(&compression, 4, bitmap_file, Endian);
		fread_endian(&imageSize, 4, bitmap_file, Endian);

		/* Skip 8 bytes of size info: X and Y pixels per meter */
		TAP_Hdd_Fseek(bitmap_file, 8, 1);

		fread_endian(&colorsUsed, 4, bitmap_file, Endian);
		TAP_Print("Colours used %d\r\n", colorsUsed);

		if (colorsUsed == 0)
			colorsUsed = (1UL << bitCount);

		fread_endian(&colorsImportant, 4, bitmap_file, Endian);

		if (colorsImportant == 0)
			colorsImportant = colorsUsed;
	}
	else if (hdrVersion != 40) {
		TAP_Hdd_Fclose(bitmap_file);
		return NULL;
	}

	// we only support 24-bit bitmaps
	if (bitCount != 24) {
		TAP_Hdd_Fclose(bitmap_file);
		return NULL;
	}

	// allocate memory for the bitmap data
	TAP_Print("Width %d\r\n", *width);
	TAP_Print("Height %d\r\n", *height);
	TAP_Print("Allocate %d for bitmap %s\r\n", *width * *height * 3, name);
	bitmap_data = (byte *) dlmalloc(*width * *height * 3);
	if (!bitmap_data) {
		TAP_Hdd_Fclose(bitmap_file);
		return NULL;
	}

	// allocate memory for the target data
	TAP_Print("Allocate %d for target\r\n", *width * *height * 2);
	target_data = (word *) dlmalloc(*width * *height * 2);
	if (!target_data) {
		dlfree(bitmap_data); bitmap_data = NULL;
		dlmalloc_trim(0);
		TAP_Hdd_Fclose(bitmap_file);
		return NULL;
	}

	// read the bitmap data
	TAP_Print("Read the file\r\n");
	TAP_Hdd_Fread(bitmap_data, *width * *height * 3, 1, bitmap_file);
	TAP_Hdd_Fclose(bitmap_file);
	
	// convert the bitmap data into display data
	TAP_Print("Convert to display format\r\n");
	convertBitmapToPixels(target_data, bitmap_data, *width, *height);

	// free the bitmap data
	dlfree(bitmap_data); bitmap_data = NULL;
	dlmalloc_trim(0);

	// finally return the screen data
	return target_data;
}

void BMP_SetPixel(byte *bitmap_buffer, int width, int height, int x, int y, byte r, byte g, byte b)
{
	int pos = ((height-y-1) * width + x) * 3;
    bitmap_buffer[pos] = b;
    bitmap_buffer[pos+1] = g;
    bitmap_buffer[pos+2] = r;
}

void captureVideo(word *bitmap_buffer, int width, int height)
{
	//TYPE_OsdBaseInfo osdBaseInfo;
	TYPE_VideoBank video_bank;

	byte *pic_address, *chr_address;
	int	x, y, xx, yy;
	byte r, g, b;
	int ypos, ypos2;
	int xpos;
	int yPos, cPos;

	memset(&video_bank, 0, sizeof(video_bank));

	TAP_CaptureScreen(CHANNEL_Main, &video_bank);

	pic_address = video_bank.yAddress;
	chr_address = video_bank.cAddress;

	//BMP_WriteHeader(file_name, width, height);
	for (y = 0; y < height; y++) 
	{
		yy = y * video_bank.height / height;
 		ypos  = ((yy    & 0xffffffe0) * 736)  + ((yy&0x1f)<<5);		//(y&0x1f)*32;
		ypos2 = (((yy/2) & 0xffffffe0) * 736)  + (((yy/2)&0x1f)<<5);	//((y/2)&0x1f)*32;
		for (x = 0; x < width; x++)
		{
			xx = x * video_bank.width / width;
			xpos =  ((xx & 0xffffffe0)<<5) + (xx&0x1f);
			yPos  = ypos  + xpos;
			cPos = (ypos2 + xpos)&0xfffffffe;

			YUV2RGB(pic_address[yPos], chr_address[cPos+1], chr_address[cPos], &r, &g, &b);

			//BMP_SetPixel(bitmap_buffer, width, height, x, y, r, g, b); // bottom -> top
			bitmap_buffer[x + (y * width)] = RGB8888(r, g, b);
		}
	}
	//BMP_End(width, height);

	TAP_MemFree(video_bank.cAddress);
	TAP_MemFree(video_bank.yAddress);
}



