#ifdef WIN32
#pragma comment(lib, "c:\\TopfieldDisk\\DataFiles\\SDK\\TAP_SDK_DLL")
#include "TAP_Emulations.h"
#endif

#include "tap.h"
#include "gif_lib.h"
#include "dlmalloc.h"

/* Make some variables global, so we could access them faster: */
static int
	ImageNum = 0,
	BackGround = 0,
	AlphaFlag = FALSE,
	HelpFlag = FALSE,
	ColorMapSize = 0,
	InterlacedOffset[] = { 0, 4, 2, 1 }, /* The way Interlaced image should. */
	InterlacedJumps[] = { 8, 8, 4, 2 };    /* be read - offsets and jumps... */
static GifColorType
	*ColorMap;

extern int _GifError;

byte *LoadGifDirect(char *name, int *bitmap_width, int *bitmap_height)
{
	int i, j, Error, NumFiles, Size, Row, Col, Width, Height, ExtCode, Count;
	GifRecordType RecordType;
	GifByteType *Extension;
	char **FileName = NULL;
	byte *ScreenBuffer;
	byte *target_data, *local_target_data;
	GifFileType *GifFile;
	dword	width, height;
	static byte lut[256][4];
    dword x, y;

	GifFile = DGifOpen( name );
	if( GifFile == NULL )
	{
		//TAP_Print(" error code = %d (%x)\r\n",    _GifError, _GifError );
		return NULL;
	}

	Size = GifFile -> SWidth;

	ScreenBuffer = (byte *)dlmalloc(Size);
	if (!ScreenBuffer)
		return NULL;

	// allocate memory for the target data
	local_target_data = target_data = (byte *) dlmalloc(GifFile->SWidth * GifFile->SHeight);
	if (!target_data) {
		dlfree(ScreenBuffer); ScreenBuffer = NULL;
		dlmalloc_trim(0);
		return NULL;
	}

	memset(local_target_data, 0, GifFile->SWidth * GifFile->SHeight);
	*bitmap_width = width = GifFile->SWidth; *bitmap_height = height = GifFile->SHeight;

	TAP_Print("width = %d, height = %d\r\n", GifFile->SWidth, GifFile->SHeight);

	BackGround = GifFile -> SBackGroundColor;
	ColorMap = (GifFile -> IColorMap ? GifFile -> IColorMap : GifFile -> SColorMap);
	ColorMapSize = 1 << (GifFile -> IColorMap ? GifFile -> IBitsPerPixel : GifFile -> SBitsPerPixel);

	for( i=0; i<256; i++ )
	{
		lut[i][0] = ColorMap[i].Red;
		lut[i][1] = ColorMap[i].Green;
		lut[i][2] = ColorMap[i].Blue;
		lut[i][3] = 0;
	}

	TAP_Osd_SetLut( 0, lut );

	/* Scan the content of the GIF file and load the image(s) in: */
	do
	{
		if( DGifGetRecordType(GifFile, &RecordType) == GIF_ERROR )
		{
			dlfree( ScreenBuffer );
			dlfree(target_data);
			dlmalloc_trim(0);
			target_data = NULL;
			DGifCloseFile(GifFile);
			return NULL;
		}

		switch (RecordType)
		{
			case IMAGE_DESC_RECORD_TYPE:
				if (DGifGetImageDesc(GifFile) == GIF_ERROR)
				{
					dlfree( ScreenBuffer );
					dlfree(target_data);
					dlmalloc_trim(0);
					target_data = NULL;
					DGifCloseFile(GifFile);
					return NULL;
				}
				Row = GifFile -> ITop; /* Image Position relative to Screen. */
				Col = GifFile -> ILeft;
				Width = GifFile -> IWidth;
				Height = GifFile -> IHeight;
				//TAP_Print("\n: Image %d at (%d, %d) [%dx%d]:     ", ++ImageNum, Col, Row, Width, Height);
				if (GifFile -> ILeft + GifFile -> IWidth > GifFile -> SWidth ||
				   GifFile -> ITop + GifFile -> IHeight > GifFile -> SHeight)
				 {
					//TAP_Print( "Image %d is not confined to screen dimension, aborted.\n");
					dlfree( ScreenBuffer );
					dlfree(target_data);
					dlmalloc_trim(0);
					target_data = NULL;
					DGifCloseFile(GifFile);
					return NULL;
				}
				if (GifFile -> IInterlace)
				{
					/* Need to perform 4 passes on the images: */
					for (Count = i = 0; i < 4; i++)
					{
						for (j = Row + InterlacedOffset[i]; j < Row + Height; j += InterlacedJumps[i])
						{
							if (DGifGetLine(GifFile, &ScreenBuffer[Col],	Width) == GIF_ERROR)
							{
								dlfree( ScreenBuffer );
								dlfree(target_data);
								dlmalloc_trim(0);
								target_data = NULL;
								DGifCloseFile(GifFile);
								return NULL;
							}
							//TAP_Osd_DrawPixmap( rgn, x, y+j, width, 1, ScreenBuffer, FALSE, OSD_256 );
							memcpy(local_target_data, ScreenBuffer, width);
							local_target_data += width;
						}
					}
				}
				else
				{
					for (i = 0; i < Height; i++)
					{
						if (DGifGetLine(GifFile, &ScreenBuffer[Col],	Width) == GIF_ERROR)
						{
							dlfree( ScreenBuffer );
							dlfree(target_data);
							dlmalloc_trim(0);
							target_data = NULL;
							DGifCloseFile(GifFile);
							return NULL;
						}
						//TAP_Osd_DrawPixmap( rgn, x, y+Row, width, 1, ScreenBuffer, FALSE, OSD_256 );
						Row++;
						memcpy(local_target_data, ScreenBuffer, width);
						local_target_data += width;
					}
				}
				break;
			case EXTENSION_RECORD_TYPE:
				/* Skip any extension blocks in file: */
				if (DGifGetExtension(GifFile, &ExtCode, &Extension) == GIF_ERROR)
				{
					dlfree( ScreenBuffer );
					dlfree(target_data);
					dlmalloc_trim(0);
					target_data = NULL;
					DGifCloseFile(GifFile);
					return NULL;
				}
				while (Extension != NULL)
				{
					if (DGifGetExtensionNext(GifFile, &Extension) == GIF_ERROR)
					{
						dlfree( ScreenBuffer );
						dlfree(target_data);
						dlmalloc_trim(0);
						target_data = NULL;
						DGifCloseFile(GifFile);
						return NULL;
					}
				}
				break;
			case TERMINATE_RECORD_TYPE:
				break;
			default:			/* Should be traps by DGifGetRecordType. */
				break;
		}

	}
	while (RecordType != TERMINATE_RECORD_TYPE);

	dlfree( ScreenBuffer );
	dlmalloc_trim(0);

	DGifCloseFile(GifFile);

	return target_data;
}

